from functools import partial
from typing import Callable

import jax
import jax.numpy as jnp
from chex import dataclass, Array, Numeric, PRNGKey, Scalar, Shape

from reinforced_lib.agents import AgentState


@dataclass
class ParticleFilterState(AgentState):
    """
    Container for the state of the particle filter agent.

    Attributes
    ----------
    positions : Array
        Positions of the particles.
    logit_weights : Array
        Unnormalized log weights of the particles.
    """

    positions: Array
    logit_weights: Array


def simple_resample(operands: tuple[ParticleFilterState, PRNGKey]) -> ParticleFilterState:
    """
    Samples new particle positions from a categorical distribution with particle weights, then sets all weights equal.

    Parameters
    ----------
    operands : tuple[ParticleFilterState, PRNGKey]
        Tuple containing the filter state and a PRNG key.

    Returns
    -------
    ParticleFilterState
        Updated filter state.
    """

    state, key = operands
    positions_idx = jax.random.categorical(key, state.logit_weights, shape=state.positions.shape)

    return ParticleFilterState(
        positions=state.positions[positions_idx],
        logit_weights=jnp.zeros_like(state.logit_weights)
    )


def effective_sample_size(state: ParticleFilterState, threshold: Scalar = 0.5) -> bool:
    r"""
    Calculates the effective sample size [9]_ (ESS). If ESS is smaller than the number of sample times threshold,
    then a resampling is necessary.

    Parameters
    ----------
    state : ParticleFilterState
        Current state of the filter.
    threshold : float, default=0.5
        Threshold value used to decide if a resampling is necessary. :math:`thr \in (0, 1)`.

    Returns
    -------
    bool
        Information whether a resampling should be performed.

    References
    ----------
    .. [9] https://en.wikipedia.org/wiki/Effective_sample_size#Weighted_samples
    """

    weights = jax.nn.softmax(state.logit_weights)
    return 1 < jnp.sum(weights ** 2) * state.positions.size * threshold


def simple_transition(state: ParticleFilterState, key: PRNGKey, scale: Scalar, *args) -> ParticleFilterState:
    r"""
    Performs simple movement of the particle positions according to a normal distribution with
    :math:`\mu = 0` and :math:`\sigma = scale`.

    Parameters
    ----------
    state : ParticleFilterState
        Current state of the filter.
    key : PRNGKey
        A PRNG key used as the random key.
    scale : float
        Scale of the random movement of particles. :math:`scale > 0`.

    Returns
    -------
    ParticleFilterState
        Updated filter state.
    """

    positions_update = scale * jax.random.normal(key, state.positions.shape)

    return ParticleFilterState(
        positions=state.positions + positions_update,
        logit_weights=state.logit_weights
    )


def linear_transition(state: ParticleFilterState, key: PRNGKey, scale: Scalar, delta_time: Scalar) -> ParticleFilterState:
    r"""
    Performs movement of the particle positions according to a normal distribution with :math:`\mu = 0` and
    :math:`\sigma = scale \cdot \Delta t`, where :math:`\Delta t` is the time elapsed since the last update.

    Parameters
    ----------
    state : ParticleFilterState
        Current state of the filter.
    key : PRNGKey
        A PRNG key used as the random key.
    scale : float
        Scale of the random movement of particles. :math:`scale > 0`.
    delta_time : float
        Time elapsed since the last update.

    Returns
    -------
    ParticleFilterState
        Updated filter state.
    """

    return simple_transition(state, key, scale * delta_time)


def affine_transition(state: ParticleFilterState, key: PRNGKey, scale: Array, delta_time: Scalar) -> ParticleFilterState:
    r"""
    Performs movement of the particle positions according to a normal distribution with :math:`\mu = 0` and
    :math:`\sigma = scale_0 \cdot \Delta t + scale_1`, where :math:`\Delta t` is the time elapsed since
    the last update.

    Parameters
    ----------
    state : ParticleFilterState
        Current state of the filter.
    key : PRNGKey
        A PRNG key used as the random key.
    scale : Array
        Scale of the random movement of particles. :math:`scale_0, scale_1 > 0`.
    delta_time : float
        Time elapsed since the last update.

    Returns
    -------
    ParticleFilterState
        Updated filter state.
    """

    return simple_transition(state, key, scale[0] * delta_time + scale[1])


class ParticleFilter:
    """
    Particle filter (sequential Monte Carlo) algorithm estimating the
    internal environment state given noisy or partial observations.

    Parameters
    ----------
    initial_distribution_fn : callable
        Function that samples the initial particle positions; takes two positional arguments:
            - ``key``: a PRNG key used as a random key (`PRNGKey`).
            - ``shape``: shape of the sample (`Shape`).

        Returns the initial particle positions (`Array`).

    positions_shape : Array
        Shape of the particle positions array.
    weights_shape : Array
        Shape of the particle weights array.
    scale : Array
        Scale of the random movement of the particles.
    observation_fn : callable
        Function that updates particles based on an observation from the environment; takes two positional arguments:
            - ``state``: the state of the filter (`ParticleFilterState`).
            - ``observation``: an observation from the environment (`any`).
        
        Returns the updated state of the filter (`ParticleFilterState`).
    
    resample_fn : callable, default=particle_filter.simple_resample
        Function that performs resampling of the particles; takes one positional argument:
            - ``operands``: a tuple containing the filter state and a PRNG key (`tuple[ParticleFilterState, PRNGKey]`).
        
        Returns the updated state of the filter (`ParticleFilterState`).
    
    resample_criterion_fn : callable, default=particle_filter.effective_sample_size
        Function that checks if a resampling is necessary; takes one positional argument:
            - ``state``: the state of the filter (`ParticleFilterState`).
        
        Returns information whether a resampling should be performed (`bool`).
    
    transition_fn : callable, default=particle_filter.simple_transition
        Function that updates the particle positions; takes four positional arguments:
            - ``state``: the state of the filter (`ParticleFilterState`).
            - ``key``: a PRNG key used as a random key (`PRNGKey`).
            - ``scale``: scale of the random movement of the particles (`Array`).
            - ``time``: the current time (`float`).
        
        Returns the updated state of the filter (`ParticleFilterState`).
    """

    def __init__(
            self,
            initial_distribution_fn: Callable,
            positions_shape: Shape,
            weights_shape: Shape,
            scale: Numeric,
            observation_fn: Callable[[ParticleFilterState, any], ParticleFilterState],
            resample_fn: Callable[[tuple[ParticleFilterState, PRNGKey]], ParticleFilterState] = simple_resample,
            resample_criterion_fn: Callable[[ParticleFilterState], bool] = effective_sample_size,
            transition_fn: Callable[[ParticleFilterState, PRNGKey, Numeric, Scalar], ParticleFilterState] = simple_transition
    ) -> None:
        self.init = jax.jit(
            partial(
                self.init,
                initial_distribution_fn=initial_distribution_fn,
                positions_shape=positions_shape,
                weights_shape=weights_shape
            )
        )
        self.update = jax.jit(
            partial(
                self.update,
                observation_fn=observation_fn,
                resample_fn=resample_fn,
                resample_criterion_fn=resample_criterion_fn,
                transition_fn=transition_fn,
                scale=scale
            )
        )
        self.sample = jax.jit(self.sample)

    @staticmethod
    def init(
            key: PRNGKey,
            initial_distribution_fn: Callable,
            positions_shape: Shape,
            weights_shape: Shape
    ) -> ParticleFilterState:
        """
        Creates and initializes an instance of the particle filter.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        initial_distribution_fn : callable
            Function that samples the initial particle positions.
                - ``key``: PRNG key used as a random key (`PRNGKey`).
                - ``shape``: shape of the sample (`Shape`).

            Returns the initial particle positions (`Array`).

        positions_shape : Array
            Shape of the particle positions array.
        weights_shape : Array
            Shape of the particle weights array.

        Returns
        -------
        ParticleFilterState
            Initial state of the Particle Filter.
        """

        return ParticleFilterState(
            positions=initial_distribution_fn(key, positions_shape),
            logit_weights=jnp.zeros(weights_shape)
        )

    @staticmethod
    def update(
            state: ParticleFilterState,
            key: PRNGKey,
            observation_fn: Callable[[ParticleFilterState, any], ParticleFilterState],
            observation: any,
            resample_fn: Callable[[tuple[ParticleFilterState, PRNGKey]], ParticleFilterState],
            resample_criterion_fn: Callable[[ParticleFilterState], bool],
            transition_fn: Callable[[ParticleFilterState, PRNGKey, Numeric, Scalar], ParticleFilterState],
            delta_time: Scalar,
            scale: Numeric
    ) -> ParticleFilterState:
        """
        Updates the state of the filter based on an observation from the environment, then
        performs resampling (if necessary) and transition of the particles.

        Parameters
        ----------
        state : ParticleFilterState
            Current state of the filter.
        key : PRNGKey
            A PRNG key used as the random key.
        observation_fn : callable
            Function that updates particles based on an observation from the environment; takes two positional arguments:
                - ``state``: the state of the filter (`ParticleFilterState`).
                - ``observation``: an observation from the environment (`any`).

            Returns the updated state of the filter (`ParticleFilterState`).
        
        observation : any
            An observation from the environment.
        resample_fn : callable, default=particle_filter.simple_resample
            Function that performs resampling of the particles; takes one positional argument:
                - ``operands``: a tuple containing the filter state and a PRNG key (`tuple[ParticleFilterState, PRNGKey]`).

            Returns the updated state of the filter (`ParticleFilterState`).

        resample_criterion_fn : callable, default=particle_filter.effective_sample_size
            Function that checks if a resampling is necessary; takes one positional argument:
                - ``state``: the state of the filter (`ParticleFilterState`).

            Returns information whether a resampling should be performed (`bool`).

        transition_fn : callable, default=particle_filter.simple_transition
            Function that updates the particle positions; takes four positional arguments:
                - ``state``: the state of the filter (`ParticleFilterState`).
                - ``key``: a PRNG key used as a random key (`PRNGKey`).
                - ``scale``: scale of the random movement of the particles (`Array`).
                - ``time``: the current time (`float`).

        delta_time : float
            Time difference between the current and the previous observation.
        scale : Array
            Scale of the random movement of the particles.

        Returns
        -------
        ParticleFilterState
            Updated filter state.
        """

        resample_key, transition_key = jax.random.split(key)

        state = observation_fn(state, observation)
        state = jax.lax.cond(resample_criterion_fn(state), resample_fn, lambda op: op[0], (state, resample_key))
        state = transition_fn(state, transition_key, scale, delta_time)

        return ParticleFilterState(
            positions=state.positions,
            logit_weights=state.logit_weights
        )

    @staticmethod
    def sample(
            state: ParticleFilterState,
            key: PRNGKey
    ) -> Numeric:
        """
        Samples the estimated environment state from a categorical distribution with particle weights.

        Parameters
        ----------
        state : ParticleFilterState
            Current state of the filter.
        key : PRNGKey
            A PRNG key used as the random key.

        Returns
        -------
        Array
            Estimated environment state.
        """

        return state.positions[jax.random.categorical(key, state.logit_weights)]



from abc import ABC
from enum import Enum
from typing import Union

from chex import Array, Scalar

from reinforced_lib.utils.exceptions import UnsupportedLogTypeError


class SourceType(Enum):
    OBSERVATION = 0
    STATE = 1
    METRIC = 2


Source = Union[tuple[str, SourceType], str, None]


class BaseLogger(ABC):
    """
    Base interface for loggers.
    """

    def __init__(self, **kwargs):
        pass

    def init(self, sources: list[Source]) -> None:
        """
        Initializes the logger given the list of all sources defined by the user.

        Parameters
        ----------
        sources : list[Source]
            List containing the sources to log.
        """

        pass

    def finish(self) -> None:
        """
        Finalizes the loggers work (e.g., closes file or shows plots).
        """

        pass

    def log_scalar(self, source: Source, value: Scalar, custom: bool) -> None:
        """
        Method of the logger interface used for logging scalar values.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        custom : bool
            Whether the source is a custom source.
        """

        raise UnsupportedLogTypeError(type(self), type(value))

    def log_array(self, source: Source, value: Array, custom: bool) -> None:
        """
        Method of the logger interface used for logging one-dimensional arrays.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        custom : bool
            Whether the source is a custom source.
        """

        raise UnsupportedLogTypeError(type(self), type(value))

    def log_dict(self, source: Source, value: dict, custom: bool) -> None:
        """
        Method of the logger interface used for logging dictionaries.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : dict
            Dictionary to log.
        custom : bool
            Whether the source is a custom source.
        """

        raise UnsupportedLogTypeError(type(self), type(value))

    def log_other(self, source: Source, value: any, custom: bool) -> None:
        """
        Method of the logger interface used for logging other values.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        custom : bool
            Whether the source is a custom source.
        """

        raise UnsupportedLogTypeError(type(self), type(value))

    @staticmethod
    def source_to_name(source: Source) -> str:
        """
        Returns a full name of the source. If source is a string itself, returns that string.
        Otherwise, it returns a string in the format "name-sourcetype" (e.g., "action-metric").

        Parameters
        ----------
        source : Source
            Source of the logged value.

        Returns
        -------
        str
            Name of the source.
        """

        if isinstance(source, tuple):
            return f'{source[0]}-{source[1].name.lower()}'
        else:
            return source



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Numeric, PRNGKey

from reinforced_lib.agents import AgentState, BaseAgent


@dataclass
class RoundRobinSchedulerState(AgentState):
    r"""
    Container for the state of the round-robin scheduler.

    Attributes
    ----------
    item : Numeric
        Scheduled item.
    """

    item: Numeric


class RoundRobinScheduler(BaseAgent):
    r"""
    Round-robin with MAB interface. This scheduler pics item sequentially.
    Sampling is deterministic, one must call ``update`` to change state.

    Parameters
    ----------
    n_arms : int
        Number of items to choose from. :math:`N \in \mathbb{N}_{+}`.
    initial_item: int, default=0
        Initial item to start sampling from.
    """

    def __init__(self, n_arms: int, initial_item: int = 0) -> None:
        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, item=initial_item))
        self.update = jax.jit(partial(self.update, n_arms=n_arms))
        self.sample = jax.jit(self.sample)

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Space:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(key: PRNGKey, item: Numeric) -> RoundRobinSchedulerState:
        return RoundRobinSchedulerState(item=jnp.asarray(item))

    @staticmethod
    def update(state: RoundRobinSchedulerState, key: PRNGKey, n_arms: int) -> RoundRobinSchedulerState:
        return RoundRobinSchedulerState(item=(state.item + 1) % n_arms)

    @staticmethod
    def sample(state: RoundRobinSchedulerState, key: PRNGKey) -> Numeric:
        return state.item



from abc import ABC, abstractmethod
from functools import wraps, partial
from typing import Callable

import gymnasium as gym
import jax
from chex import dataclass, PRNGKey, ArrayTree

from reinforced_lib.utils.exceptions import UnimplementedSpaceError
from reinforced_lib.utils import is_array, is_dict


@dataclass
class AgentState:
    """
    Base class for agent state containers.
    """


class BaseAgent(ABC):
    """
    Base interface of agents.
    """

    @staticmethod
    @abstractmethod
    def init(key: PRNGKey, *args, **kwargs) -> AgentState:
        """
        Creates and initializes instance of the agent.
        """

        pass

    @staticmethod
    @abstractmethod
    def update(state: AgentState, key: PRNGKey, *args, **kwargs) -> AgentState:
        """
        Updates the state of the agent after performing some action and receiving a reward.
        """

        pass

    @staticmethod
    @abstractmethod
    def sample(state: AgentState, key: PRNGKey, *args, **kwargs) -> any:
        """
        Selects the next action based on the current environment and agent state.
        """

        pass

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        """
        Parameters of the agent constructor in Gymnasium format. Type of returned value is required  to
        be ``gym.spaces.Dict`` or ``None``. If ``None``, the user must provide all parameters manually.
        """

        return None

    @property
    def update_observation_space(self) -> gym.spaces.Space:
        """
        Observation space of the ``update`` method in Gymnasium format. Allows to infer missing
        observations using an extensions and easily export the agent to TensorFlow Lite format.
        If ``None``, the user must provide all parameters manually.
        """

        return None

    @property
    def sample_observation_space(self) -> gym.spaces.Space:
        """
        Observation space of the ``sample`` method in Gymnasium format. Allows to infer missing
        observations using an extensions and easily export the agent to TensorFlow Lite format.
        If ``None``, the user must provide all parameters manually.
        """

        return None

    @property
    def action_space(self) -> gym.spaces.Space:
        """
        Action space of the agent in Gymnasium format.
        """

        raise NotImplementedError()

    def export(self, init_key: PRNGKey, state: AgentState = None, sample_only: bool = False) -> tuple[any, any, any]:
        """
        Exports the agent to TensorFlow Lite format.

        Parameters
        ----------
        init_key : PRNGKey
            Key used to initialize the agent.
        state : AgentState, optional
            State of the agent to be exported. If not specified, the agent is initialized with ``init_key``.
        sample_only : bool, optional
            If ``True``, the exported agent will only be able to sample actions, but not update its state.
        """

        try:
            import tensorflow as tf
        except ModuleNotFoundError:
            raise ModuleNotFoundError('TensorFlow installation is required to export agents to TensorFlow Lite.')

        @dataclass
        class TfLiteState:
            state: ArrayTree
            key: PRNGKey

        def append_value(value: any, value_name: str, args: any) -> any:
            if args is None:
                raise UnimplementedSpaceError()
            elif is_dict(args):
                return {value_name: value} | args
            elif is_array(args):
                return [value] + list(args)
            else:
                return [value, args]

        def flatten_args(tree_args_fun: Callable, treedef: ArrayTree) -> Callable:
            @wraps(tree_args_fun)
            def flat_args_fun(*leaves):
                tree_args = jax.tree_util.tree_unflatten(treedef, leaves)

                if is_dict(tree_args):
                    tree_ret = tree_args_fun(**tree_args)
                else:
                    tree_ret = tree_args_fun(*tree_args)

                return jax.tree_util.tree_leaves(tree_ret)

            return flat_args_fun

        def make_converter(fun: Callable, arguments: any) -> tf.lite.TFLiteConverter:
            leaves, treedef = jax.tree_util.tree_flatten(arguments)
            flat_fun = flatten_args(fun, treedef)

            inputs = [[(f'arg{i}', l) for i, l in enumerate(leaves)]]
            converter = tf.lite.TFLiteConverter.experimental_from_jax([flat_fun], inputs)
            converter.target_spec.supported_ops = [
                tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
                tf.lite.OpsSet.SELECT_TF_OPS     # enable TensorFlow ops.
            ]

            return converter

        def init() -> TfLiteState:
            return TfLiteState(
                state=self.init(init_key),
                key=init_key
            )

        def sample(state: TfLiteState, *args, **kwargs) -> tuple[any, TfLiteState]:
            sample_key, key = jax.random.split(state.key)
            action = self.sample(state.state, sample_key, *args, **kwargs)
            return action, TfLiteState(state=state.state, key=key)

        def update(state: TfLiteState, *args, **kwargs) -> TfLiteState:
            update_key, key = jax.random.split(state.key)
            new_state = self.update(state.state, update_key, *args, **kwargs)
            return TfLiteState(state=new_state, key=key)

        def get_key() -> PRNGKey:
            return init_key

        def sample_without_state(state: AgentState, key: PRNGKey, *args, **kwargs) -> tuple[any, PRNGKey]:
            sample_key, key = jax.random.split(key)
            action = self.sample(state, sample_key, *args, **kwargs)
            return action, key

        if not sample_only:
            if state is None:
                state = init()
            else:
                state = TfLiteState(state=state, key=init_key)

            update_args = append_value(state, 'state', self.update_observation_space.sample())
            sample_args = append_value(state, 'state', self.sample_observation_space.sample())

            init_tfl = make_converter(init, []).convert()
            update_tfl = make_converter(update, update_args).convert()
            sample_tfl = make_converter(sample, sample_args).convert()

            return init_tfl, update_tfl, sample_tfl
        elif state is not None:
            sample_args = append_value(init_key, 'key', self.sample_observation_space.sample())

            init_tfl = make_converter(get_key, []).convert()
            sample_tfl = make_converter(partial(sample_without_state, state), sample_args).convert()

            return init_tfl, None, sample_tfl
        else:
            raise ValueError('Either `state` must be provided or `sample_only` must be False.')



from collections import defaultdict

import wandb
from chex import Array, Scalar

from reinforced_lib.logs import BaseLogger, Source


class WeightsAndBiasesLogger(BaseLogger):
    r"""
    Logger that saves values to Weights & Biases [4]_ platform. ``WeightsAndBiasesLogger`` synchronizes
    the logged values in time. This means that if the same source is logged less often than other sources,
    the step will be increased accordingly to maintain the appropriate spacing between the values on the x-axis.

    **Note**: to use this logger, you need to log into W&B before running the script. The necessary steps are
    described in the official documentation [4]_.

    Parameters
    ----------
    wandb_sync_steps : bool, default=False
        Set to ``True`` if you want to synchronize the logged values in time.
    wandb_kwargs : dict, optional
        Additional keyword arguments passed to ``wandb.init`` function.

    References
    ----------
    .. [4] Weights & Biases. https://docs.wandb.ai/
    """

    def __init__(
            self,
            wandb_sync_steps: bool = False,
            wandb_kwargs: dict = None,
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self._sync_steps = wandb_sync_steps
        self._current_values = set()
        self._step = 0
        self._steps = defaultdict(int)

        wandb.init(**(wandb_kwargs or {}))
        wandb.define_metric('*', step_metric='global_step')

    def finish(self) -> None:
        """
        Finishes the W&B run.
        """

        wandb.finish()

    def log_scalar(self, source: Source, value: Scalar, *_) -> None:
        """
        Logs a scalar value to the W&B logger.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : float
            Scalar to log.
        """

        self._log(source, value)

    def log_array(self, source: Source, value: Array, *_) -> None:
        """
        Logs an array to the W&B logger.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Array
            Array to log.
        """

        self._log(source, value)

    def log_dict(self, source: Source, value: dict, *_) -> None:
        """
        Logs a dictionary to the W&B logger.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : dict
            Dictionary to log.
        """

        self._log(source, value)

    def log_other(self, source: Source, value: any, *_) -> None:
        """
        Logs an object to the W&B logger.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : any
            Value of any type to log.
        """

        self._log(source, value)

    def _log(self, source: Source, value: any) -> None:
        """
        Adds a given value to the logger.

        Parameters
        ----------
        source : Source
            Source of the logged value.
        value : Numeric
            Value to log.
        """

        name = self.source_to_name(source)
        step = self._get_step(name)
        wandb.log({'global_step': step, name: value})

    def _get_step(self, name: str) -> int:
        """
        Returns the current step for a given source.

        Parameters
        ----------
        name : str
            Name of the source.

        Returns
        -------
        int
            Current step for the given source.
        """

        if self._sync_steps:
            if name in self._current_values:
                self._step += 1
                self._current_values.clear()

            self._current_values.add(name)
            step = self._step
        else:
            step = self._steps[name] + 1

        self._steps[name] = step
        return step



from typing import Callable

import jax
import jax.numpy as jnp
from chex import dataclass, Array, Numeric, Scalar, PRNGKey, Shape


@dataclass
class ExperienceReplay:
    """
    Container for experience replay buffer functions.

    Attributes
    ----------
    init : Callable
        Function that initializes the replay buffer.
    append : Callable
        Function that appends a new values to the replay buffer.
    sample : Callable
        Function that samples a batch from the replay buffer.
    is_ready : Callable
        Function that checks if the replay buffer is ready to be sampled.
    """

    init: Callable
    append: Callable
    sample: Callable
    is_ready: Callable


@dataclass
class ReplayBuffer:
    """
    Dataclass containing the replay buffer values. The replay buffer is implemented as a circular buffer.

    Attributes
    ----------
    states : Array
        Array containing the states.
    actions : Array
        Array containing the actions.
    rewards : Array
        Array containing the rewards.
    terminals : Array
        Array containing the terminal flags.
    next_states : Array
        Array containing the next states.
    size : int
        Current size of the replay buffer.
    ptr : int
        Current pointer of the replay buffer.
    """

    states: Array
    actions: Array
    rewards: Array
    terminals: Array
    next_states: Array
    size: int
    ptr: int


def experience_replay(
        buffer_size: int,
        batch_size: int,
        obs_space_shape: Shape,
        act_space_shape: Shape
) -> ExperienceReplay:
    """
    Experience replay buffer used for off-policy learning. Improves the stability of the learning process
    by reducing the correlation between the samples and enables an agent to learn from past experiences.

    Parameters
    ----------
    buffer_size : int
        Maximum size of the replay buffer.
    batch_size : int
        Size of the batch to be sampled from the replay buffer.
    obs_space_shape : Shape
        Shape of the observation space.
    act_space_shape : Shape
        Shape of the action space.

    Returns
    -------
    out : ExperienceReplay
        Container for experience replay buffer functions.
    """

    def init() -> ReplayBuffer:
        """
        Initializes the replay buffer with empty arrays.

        Returns
        -------
        buffer : ReplayBuffer
            Dataclass containing the replay buffer values.
        """

        return ReplayBuffer(
            states=jnp.empty((buffer_size, *obs_space_shape)),
            actions=jnp.empty((buffer_size, *act_space_shape)),
            rewards=jnp.empty((buffer_size, 1)),
            terminals=jnp.empty((buffer_size, 1), dtype=jnp.bool_),
            next_states=jnp.empty((buffer_size, *obs_space_shape)),
            size=0,
            ptr=0
        )

    def append(
            buffer: ReplayBuffer,
            state: Numeric,
            action: Numeric,
            reward: Scalar,
            terminal: bool,
            next_state: Numeric
    ) -> ReplayBuffer:
        """
        Appends a new values to the replay buffer.

        Parameters
        ----------
        buffer : ReplayBuffer
            Dataclass containing the replay buffer values.
        state : Array
            State of the environment.
        action : Array
            Action taken by the agent.
        reward : float
            Reward received by the agent.
        terminal : bool
            Flag indicating if the episode has terminated.
        next_state : Array
            Next state of the environment.

        Returns
        -------
        buffer : ReplayBuffer
            Updated replay buffer.
        """

        return ReplayBuffer(
            states=buffer.states.at[buffer.ptr].set(state),
            actions=buffer.actions.at[buffer.ptr].set(action),
            rewards=buffer.rewards.at[buffer.ptr].set(reward),
            terminals=buffer.terminals.at[buffer.ptr].set(terminal),
            next_states=buffer.next_states.at[buffer.ptr].set(next_state),
            size=jax.lax.min(buffer.size + 1, buffer_size),
            ptr=(buffer.ptr + 1) % buffer_size
        )

    def sample(buffer: ReplayBuffer, key: PRNGKey) -> tuple:
        """
        Samples a batch from the replay buffer (there may be duplicates!).

        Parameters
        ----------
        buffer : ReplayBuffer
            Dataclass containing the replay buffer values.
        key : PRNGKey
            Pseudo-random number generator key.

        Returns
        -------
        batch : tuple
            Tuple containing the batch of states, actions, rewards, terminals and next states.
        """

        idxs = jax.random.uniform(key, shape=(batch_size,), minval=0, maxval=buffer.size).astype(int)

        states = buffer.states[idxs]
        actions = buffer.actions[idxs]
        rewards = buffer.rewards[idxs]
        terminals = buffer.terminals[idxs]
        next_states = buffer.next_states[idxs]

        return states, actions, rewards, terminals, next_states

    def is_ready(buffer: ReplayBuffer) -> bool:
        """
        Checks if the replay buffer is ready to be sampled (contains at least ``batch_size`` elements).

        Parameters
        ----------
        buffer : ReplayBuffer
            Dataclass containing the replay buffer values.

        Returns
        -------
        ready : bool
            Flag indicating if the replay buffer is ready to be sampled.
        """

        return buffer.size >= batch_size

    return ExperienceReplay(
        init=jax.jit(init),
        append=jax.jit(append),
        sample=jax.jit(sample),
        is_ready=jax.jit(is_ready)
    )



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, PRNGKey

from reinforced_lib.agents import AgentState, BaseAgent


@dataclass
class RandomSchedulerState(AgentState):
    r"""Random scheduler has no memory, thus the state is empty."""
    pass


class RandomScheduler(BaseAgent):
    r"""
    Random scheduler with MAB interface. This scheduler pics item randomly.

    Parameters
    ----------
    n_arms : int
        Number of items to choose from. :math:`N \in \mathbb{N}_{+}`.
    """

    def __init__(self, n_arms: int) -> None:
        self.n_arms = n_arms

        self.init = jax.jit(self.init)
        self.update = jax.jit(self.update)
        self.sample = jax.jit(partial(self.sample, n_arms=n_arms))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Space:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(key: PRNGKey) -> RandomSchedulerState:
        return RandomSchedulerState()

    @staticmethod
    def update(state: RandomSchedulerState, key: PRNGKey) -> RandomSchedulerState:
        return state

    @staticmethod
    def sample(state: RandomSchedulerState, key: PRNGKey, n_arms: int) -> int:
        return jax.random.choice(key, n_arms)



from reinforced_lib.exts.utils import observation, parameter
from reinforced_lib.exts.base_ext import BaseExt
from reinforced_lib.exts.basic_mab import BasicMab
from reinforced_lib.exts.gymnasium import Gymnasium



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, Scalar, PRNGKey

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class Exp3State(AgentState):
    """
    Container for the state of the Exp3 agent.

    Attributes
    ----------
    omega : Array
        Preference for each arm.
    """

    omega: Array


class Exp3(BaseAgent):
    r"""
    Basic Exp3 agent for stationary multi-armed bandit problems with exploration factor :math:`\gamma`. The higher
    the value, the more the agent explores. The implementation is inspired by the work of Auer et al. [6]_. There
    are many variants of the Exp3 algorithm, you can find more information in the original paper.

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}`.
    gamma : float
        Exploration factor. :math:`\gamma \in (0, 1]`.
    min_reward : float
        Minimum possible reward.
    max_reward : float
        Maximum possible reward.

    References
    ----------
    .. [6] Peter Auer, Nicolò Cesa-Bianchi, Yoav Freund, and Robert E. Schapire. 2002. The Nonstochastic Multiarmed
       Bandit Problem. SIAM Journal on Computing, 32(1), 48–77.
    """

    def __init__(
            self,
            n_arms: int,
            gamma: Scalar,
            min_reward: Scalar,
            max_reward: Scalar
    ) -> None:
        assert 0 < gamma <= 1

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=n_arms))
        self.update = jax.jit(partial(self.update, gamma=gamma, min_reward=min_reward, max_reward=max_reward))
        self.sample = jax.jit(partial(self.sample, gamma=gamma))

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'gamma': gym.spaces.Box(0, 1, (1,), float),
            'min_reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float),
            'max_reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(
            key: PRNGKey,
            n_arms: int
    ) -> Exp3State:
        """
        Initializes the Exp3 agent state with uniform preference for each arm.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.

        Returns
        -------
        Exp3State
            Initial state of the Exp3 agent.
        """

        return Exp3State(
            omega=jnp.ones((n_arms, 1)) / n_arms
        )

    @staticmethod
    def update(
        state: Exp3State,
        key: PRNGKey,
        action: int,
        reward: Scalar,
        gamma: Scalar,
        min_reward: Scalar,
        max_reward: Scalar
    ) -> Exp3State:
        r"""
        Agent updates its preference for the selected arm :math:`a` according to the following formula:

        .. math::
            \omega_{t + 1}(a) = \omega_{t}(a) \exp \left( \frac{\gamma r}{\pi(a) K} \right)

        where :math:`\omega_{t + 1}(a)` is the preference of arm :math:`a` at time :math:`t + 1`, :math:`\pi(a)` is
        the probability of selecting arm :math:`a`, and :math:`K` is the number of arms. The reward :math:`r` is
        normalized to the range :math:`[0, 1]`. The exponential growth significantly increases the weight of good arms,
        so in the long use of the agent it is important to **ensure that the values of** :math:`\omega` **do not exceed
        the maximum value of the floating point type!**

        Parameters
        ----------
        state : Exp3State
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : float
            Reward collected by the agent after taking the previous action.
        gamma : float
            Exploration factor.
        min_reward : float
            Minimum possible reward.
        max_reward : float
            Maximum possible reward.

        Returns
        -------
        Exp3State
            Updated agent state.
        """

        reward = (reward - min_reward) / (max_reward - min_reward)
        n_arms = state.omega.size

        pi = (1 - gamma) * state.omega / state.omega.sum() + gamma / n_arms

        return Exp3State(
            omega=state.omega.at[action].mul(jnp.exp(gamma * reward / (pi[action] * n_arms)))
        )

    @staticmethod
    def sample(
        state: Exp3State,
        key: PRNGKey,
        gamma: Scalar
    ) -> int:
        r"""
        The Exp3 policy is stochastic. Algorithm chooses a random arm with probability :math:`\gamma`, otherwise it
        draws arm :math:`a` with probability :math:`\omega(a) / \sum_{b=1}^N \omega(b)`.

        Parameters
        ----------
        state : Exp3State
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        gamma : float
            Exploration factor.

        Returns
        -------
        int
            Selected action.
        """

        pi = (1 - gamma) * state.omega / state.omega.sum() + gamma / state.omega.size
        return jax.random.categorical(key, jnp.log(pi.flatten()))



from typing import Any, Callable

import jax
import optax
from chex import Array, Scalar
from flax import linen as nn


def gradient_step(
        objective: any,
        loss_params: tuple,
        opt_state: optax.OptState,
        optimizer: optax.GradientTransformation,
        loss_fn: Callable
) -> tuple[any, any, optax.OptState, Scalar]:
    r"""
    Performs a gradient step on the objective with respect to ``grad_loss_fn`` function.
    ``grad_loss_fn`` should return tuple of ``(loss, aux)`` where loss is the value to be minimized
    and aux is auxiliary value to be returned (can be ``None``).

    Parameters
    ----------
    objective : any
        Objective to be optimized.
    loss_params : tuple
        Parameters to pass to ``loss_fn``.
    opt_state : optax.OptState
        Optimizer state.
    optimizer : optax.GradientTransformation
        Optimizer to use for gradient step.
    loss_fn : Callable
        Function that returns the loss to be minimized. Can return additional values as well.

    Returns
    -------
    out : tuple[any, any, optax.OptState, Scalar]
        Tuple containing the updated objective and optimizer state, as well as the loss value.
    """

    (loss, aux), grads = jax.value_and_grad(loss_fn, has_aux=True)(objective, *loss_params)
    updates, opt_state = optimizer.update(grads, opt_state)
    objective = optax.apply_updates(objective, updates)

    return objective, aux, opt_state, loss


def init(model: nn.Module, key: jax.random.PRNGKey, *x: Any) -> tuple[dict, dict]:
    r"""
    Initializes the ``flax`` model.

    Parameters
    ----------
    model : nn.Module
        Model to be initialized.
    key : PRNGKey
        A PRNG key used as the random key.
    x : any
        Input to the model.

    Returns
    -------
    tuple[dict, dict]
        Tuple containing the parameters and the state of the model.
    """

    params_key, rlib_key, dropout_key = jax.random.split(key, 3)

    variables = model.init({'params': params_key, 'rlib': rlib_key, 'dropout': dropout_key}, *x)
    params = variables.pop('params')
    state = variables

    return params, state


def forward(model: nn.Module, params: dict, state: dict, key: jax.random.PRNGKey, *x: Any) -> tuple[Array, dict]:
    r"""
    Forward pass through the ``flax`` model. **Note**: by default, the model is provided with two random key
    streams: one for the dropout layers and one for the user. This is done to ensure that the dropout is always
    initialized with the same random key, and that the user can use the custom key for any other purpose.
    The custom key is available in the model by calling ``self.make_rng('rlib')``.

    Parameters
    ----------
    model : nn.Module
        Model to be used for forward pass.
    params : dict
        Parameters of the model.
    state : dict
        State of the network.
    key : PRNGKey
        A PRNG key used as the random key.
    x : any
        Input to the model.

    Returns
    -------
    tuple[Array, dict]
        Tuple containing the output of the model and the updated state.
    """

    rlib_key, dropout_key = jax.random.split(key)
    return model.apply({'params': params, **state}, *x, rngs={'rlib': rlib_key, 'dropout': dropout_key}, mutable=list(state.keys()))



from functools import partial

import gymnasium as gym
import jax
import jax.numpy as jnp
from chex import dataclass, Array, Scalar, PRNGKey

from reinforced_lib.agents import BaseAgent, AgentState


@dataclass
class EGreedyState(AgentState):
    r"""
    Container for the state of the :math:`\epsilon`-greedy agent.

    Attributes
    ----------
    Q : Array
        Action-value function estimates for each arm.
    N : Array
        Number of tries for each arm.
    e : float
        Experiment rate (epsilon).
    """

    Q: Array
    N: Array
    e: Scalar


class EGreedy(BaseAgent):
    r"""
    Epsilon-greedy [5]_ agent with an optimistic start behavior and optional exponential recency-weighted average update.
    It selects a random action from a set of all actions :math:`\mathscr{A}` with probability
    :math:`\epsilon` (exploration), otherwise it chooses the currently best action (exploitation).
    Epsilon can be decayed over time to shift the policy from exploration to exploitation.

    Parameters
    ----------
    n_arms : int
        Number of bandit arms. :math:`N \in \mathbb{N}_{+}`.
    e : float
        Initial experiment rate (epsilon). :math:`\epsilon \in [0, 1]`.
    e_min : float, default=0.0
        Minimum value of the experiment rate. :math:`\epsilon_{\min} \in [0, 1]`.
    e_decay : float, default=1.0
        Decay factor for the experiment rate. :math:`\epsilon_{\text{decay}} \in [0, 1]`.
    optimistic_start : float, default=0.0
        Interpreted as the optimistic start to encourage exploration in the early stages.
    alpha : float, default=0.0
        If non-zero, exponential recency-weighted average is used to update :math:`Q` values. :math:`\alpha \in [0, 1]`.

    References
    ----------
    .. [5] Richard Sutton and Andrew Barto. 2018. Reinforcement Learning: An Introduction. The MIT Press.
    """

    def __init__(
            self,
            n_arms: int,
            e: Scalar,
            e_min: Scalar = 0.0,
            e_decay: Scalar = 1.0,
            optimistic_start: Scalar = 0.0,
            alpha: Scalar = 0.0
    ) -> None:
        assert 0 <= e <= 1
        assert 0 <= e_min <= 1
        assert 0 <= e_decay <= 1
        assert 0 <= alpha <= 1

        self.n_arms = n_arms

        self.init = jax.jit(partial(self.init, n_arms=n_arms, e=e, optimistic_start=optimistic_start))
        self.update = jax.jit(partial(self.update, alpha=alpha, e_decay=e_decay, e_min=e_min))
        self.sample = jax.jit(self.sample)

    @staticmethod
    def parameter_space() -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'n_arms': gym.spaces.Box(1, jnp.inf, (1,), int),
            'e': gym.spaces.Box(0.0, 1.0, (1,), float),
            'e_min': gym.spaces.Box(0.0, 1.0, (1,), float),
            'e_decay': gym.spaces.Box(0.0, 1.0, (1,), float),
            'optimistic_start': gym.spaces.Box(0.0, jnp.inf, (1,), float),
            'alpha': gym.spaces.Box(0.0, 1.0, (1,), float)
        })

    @property
    def update_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({
            'action': gym.spaces.Discrete(self.n_arms),
            'reward': gym.spaces.Box(-jnp.inf, jnp.inf, (1,), float)
        })

    @property
    def sample_observation_space(self) -> gym.spaces.Dict:
        return gym.spaces.Dict({})

    @property
    def action_space(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(self.n_arms)

    @staticmethod
    def init(
            key: PRNGKey,
            n_arms: int,
            e: Scalar,
            optimistic_start: Scalar
    ) -> EGreedyState:
        r"""
        Creates and initializes instance of the :math:`\epsilon`-greedy agent for ``n_arms`` arms. Action-value function estimates are
        set to ``optimistic_start`` value and the number of tries is one for each arm.

        Parameters
        ----------
        key : PRNGKey
            A PRNG key used as the random key.
        n_arms : int
            Number of bandit arms.
        e : float
            Initial experiment rate (epsilon).
        optimistic_start : float
            Interpreted as the optimistic start to encourage exploration in the early stages.

        Returns
        -------
        EGreedyState
            Initial state of the :math:`\epsilon`-greedy agent.
        """

        return EGreedyState(
            Q=jnp.full((n_arms, 1), optimistic_start),
            N=jnp.ones((n_arms, 1), dtype=int),
            e=e
        )

    @staticmethod
    def update(
        state: EGreedyState,
        key: PRNGKey,
        action: int,
        reward: Scalar,
        alpha: Scalar,
        e_decay: Scalar,
        e_min: Scalar
    ) -> EGreedyState:
        r"""
        In the stationary case, the action-value estimate for a given arm is updated as
        :math:`Q_{t + 1} = Q_t + \frac{1}{t} \lbrack R_t - Q_t \rbrack` after receiving reward :math:`R_t` at step
        :math:`t` and the number of tries for the corresponding arm is incremented. In the non-stationary case,
        the update follows the equation :math:`Q_{t + 1} = Q_t + \alpha \lbrack R_t - Q_t \rbrack`.
        Exploration rate :math:`\epsilon` can be decayed over time following the equation
        :math:`\epsilon_{t + 1} = \max(\epsilon_{\min}, \epsilon_t \times \epsilon_{\text{decay}})`.

        Parameters
        ----------
        state : EGreedyState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.
        action : int
            Previously selected action.
        reward : float
            Reward collected by the agent after taking the previous action.
        alpha : float
            Exponential recency-weighted average factor (used when :math:`\alpha > 0`).
        e_decay : float
            Decay factor for the experiment rate.
        e_min : float
            Minimum value of the experiment

        Returns
        -------
        EGreedyState
            Updated agent state.
        """

        def classic_update(operands: tuple) -> EGreedyState:
            state, action, reward, alpha = operands
            return EGreedyState(
                Q=state.Q.at[action].add((reward - state.Q[action]) / state.N[action]),
                N=state.N.at[action].add(1),
                e=jnp.maximum(e_min, state.e * e_decay)
            )

        def erwa_update(operands: tuple) -> EGreedyState:
            state, action, reward, alpha = operands
            return EGreedyState(
                Q=state.Q.at[action].add(alpha * (reward - state.Q[action])),
                N=state.N.at[action].add(1),
                e=jnp.maximum(e_min, state.e * e_decay)
            )

        return jax.lax.cond(alpha == 0, classic_update, erwa_update, (state, action, reward, alpha))

    @staticmethod
    def sample(
        state: EGreedyState,
        key: PRNGKey
    ) -> int:
        r"""
        Epsilon-greedy agent follows the policy:

        .. math::
           A =
           \begin{cases}
              \operatorname*{argmax}_{a \in \mathscr{A}} Q(a) & \text{with probability } 1 - \epsilon , \\
              \text{random action} & \text{with probability } \epsilon . \\
           \end{cases}

        Parameters
        ----------
        state : EGreedyState
            Current state of the agent.
        key : PRNGKey
            A PRNG key used as the random key.

        Returns
        -------
        int
            Selected action.
        """

        max_q = (state.Q == state.Q.max()).astype(float)
        probs = (1 - state.e) * max_q / jnp.sum(max_q) + state.e / state.Q.shape[0]
        return jax.random.choice(key, state.Q.shape[0], p=probs.flatten())



from datetime import datetime

import numpy as np


def is_scalar(x: any) -> bool:
    """
    Checks whether the input is a scalar.

    Parameters
    ----------
    x : any
        Input to check.

    Returns
    -------
    bool
        ``True`` if the input is a scalar, ``False`` otherwise.
    """

    return np.isscalar(x) or (hasattr(x, 'ndim') and x.ndim == 0)


def is_array(x: any) -> bool:
    """
    Checks whether the input is an array.

    Parameters
    ----------
    x : any
        Input to check.

    Returns
    -------
    bool
        ``True`` if the input is an array, ``False`` otherwise.
    """

    return isinstance(x, (list, tuple)) or (hasattr(x, 'ndim') and x.ndim == 1)


def is_tensor(x: any) -> bool:
    """
    Checks whether the input is a tensor.

    Parameters
    ----------
    x : any
        Input to check.

    Returns
    -------
    bool
        ``True`` if the input is a tensor, ``False`` otherwise.
    """

    return hasattr(x, 'ndim') and x.ndim > 1


def is_dict(x: any) -> bool:
    """
    Checks whether the input is a dictionary.

    Parameters
    ----------
    x : any
        Input to check.

    Returns
    -------
    bool
        ``True`` if the input is a dictionary, ``False`` otherwise.
    """

    return isinstance(x, dict)


def timestamp() -> str:
    """
    Returns the current timestamp.

    Returns
    -------
    str
        Current timestamp.
    """

    return datetime.now().strftime('%Y%m%d-%H%M%S')
